#ifndef CW_k_UNIVERSAL_H_
#define CW_k_UNIVERSAL_H_

#include <HashFunction.h>

/**
 * Defines a class of k independent Hashfunctions.
 * The implementation is based on the CW-Trick shown in:
 * "Tabulation Based 5-Universal Hashing and Linear Probing",
 * M.Thorup, Y.Zhang, SIAM
 * The k can be chosen freely and realises a hashfunction: (a_(k-1) * x^(k-1) +...+ a_1 * x + a_0)%p %m.
 */

class CWkUniversal: public HashFunction
{
public:
      
	/**
	 * The two hash functions.
	 * @param _key defines the number used for calculating the hash value.
	 */
      	sizetype h1(keytype _key);
      	sizetype h2(keytype _key);
      	
	/**
	 * Constructor commonly called by the create function.
	 * @param _hashsize defines the size of the hashtable in which to operate.
	 * @param _universality defines the k
	 */
	CWkUniversal(sizetype _hashsize, uint32_t _universality);

	/**
	 * Destructor.
	 */
      	~CWkUniversal();
	
	/**
	 * Defines the maximum operatable key.
	 * @return the maximum operatable key.
	 */
      	universetype getmaxUniverse(){return p_prime-1;};
	
	/**
	 * Shows all parameters used by this function.
	 * @param _cnt will contain the length of the array after execution.
	 * @returns an array containing all parameters.
	 */
      	universetype* getParam(uint32_t& _cnt){
      	  universetype * res = new universetype[12];
	  _cnt=2*p_k+2;
	  for(uint32_t i=0; i<p_k;i++)
	  {
		res[i]=p_param1[i];
		res[i+p_k]=p_param2[i];
	  }
	  res[2*p_k]=p_prime;
	  res[2*p_k+1]=p_m;
	  return res;
      	}

	/**
	 * defines the name of the Hashfunction.
	 * @returns a C-String containing the name of the hashfunction.
	 */
      	char* getSpecifier(){return const_cast<char*>("CW-Trick k Universal");};
 
private:
 	/**
	 * Arrays containing the a parameters for the hashfunctions.
	 */
      	universetype* p_param1;
      	universetype* p_param2;

	/**
	 * Parameters used by both Hashfunctions in common.
	 */
      	universetype p_prime, p_m;

	/**
	 * contains the degree of independence of this Hashfunction
	 */
      	uint32_t p_k;
      
	/**
	 * The implementation of the CW-Trick. Calculates (a*x+b)%p + j*p, j in {0,1}.
	 * @param x the x value of (a*x+b)
	 * @param a the a value of (a*x+b)
	 * @param b the b value of (a*x+b)
	 * @return (a*x+b)%p + j*p , j in {0,1}
	 */
      	inline universetype CWTrick(keytype x, universetype a, universetype b);      
};

#endif
