#include "CuckooHashing.h"

#include <iostream>
#include <cmath>

HashTable::HashTable(sizetype _size, sizetype _stashsize, keytype _universe,  int _functiontype, int _looptype, float _epsilon)
{
	//Sets the tablesize according to the defined epsilon.
	p_tablesize = _size * (1+_epsilon);

	//stores the unviersesize
	p_universe = _universe;

	// creates both tables used for storing the keys.
	p_hashtable1 = new keytype[p_tablesize];
	p_hashtable2 = new keytype[p_tablesize];
	
	// initialize all cells with zero
	for(sizetype i=0; i<p_tablesize;i++)
	{
	    p_hashtable1[i]=0;
	    p_hashtable2[i]=0;
	}
	
	//initialize the amount of keys on the stash as zero.
	p_stashheight = 0;
	///store the maximum size of the stash
	p_stashsize = _stashsize;
	//initialize the stash.
	p_stash = new keytype[_stashsize];

	
	sizetype stash=p_stashsize;
	double ml;	
	
	//initialize the maximal number of iterations according to the loop type
	switch(_looptype)
	{
	case SHORTLOOP:
		ml = 3.0*((double)stash + 2.0)* log((double)_size)/log(1.0+_epsilon);
		p_max_loop = (uint32_t)ml;
		p_max_loop = (ml>p_max_loop)?p_max_loop+1:p_max_loop;
		break;
	case LONGLOOP:
	default:
		p_max_loop=2*p_tablesize+3;
		break;
	}
	
	//create the 2 functions that are going to be used according to the chosen parameters
	p_function = HashFunction::create(_functiontype, p_tablesize, stash);
}
HashTable::~HashTable()
{
  // all allocated memory for the hashtables are freed
    delete[] p_hashtable1;
    p_hashtable1=0;
    delete[] p_hashtable2;
    p_hashtable2=0;
    delete[] p_stash;
    p_stash=0;
    delete p_function;
    p_function = 0;
}

uint64_t HashTable::insert(keytype _key)
{
	//initialize the number of hashfunction calculations for this insert
	uint64_t calcs = 0;
	keytype nestless = _key;
	keytype swap=_key;
	sizetype cell;
	//try to fit the key into one of both hashtables
	for(sizetype i=0;i<p_max_loop;i++)
	{
		///increase the number of calculations
		calcs++;
		///binary and returns true if the least significant bit is 1, if it is an uneven number
		if(i&1){
			//try to fit the nestless element into the second hashtable
			//calculate the position of the element in the second hashtable
			cell=p_function->h2(nestless);
			//read the element at this position
			swap = p_hashtable2[cell];
			//store the nestless element at this position
			p_hashtable2[cell] = nestless;
			// set the read element as new nestless element
			nestless=swap;
		}
		else{
			//try to fit the nestless element into the first hashtable
			//calculate the position of the element in the first hashtable
			cell=p_function->h1(nestless);
			//read the element currently at this position
			swap = p_hashtable1[cell];
			// store the nestless element at this position
			p_hashtable1[cell] = nestless;
			//set the read element as the new nestless
			nestless=swap;
		}
		//if the cell in the hashtable was empty the key was succesfull inserted
		if(nestless==0)
		    return calcs;
	}
	
	//the maximal number of loops has been reached and the nestless element has not been zero

	//check if the stash is not already full
	if(p_stashheight < p_stashsize){
		//store the nestless element in the stash
		p_stash[p_stashheight]=nestless;
		//increase stashheight
		p_stashheight++;
		return calcs;
	}else{
		//doesn't do anything
		rehash();
		//return that a rehash occured
		return 0;
	}
}

void HashTable::rehash()
{
	//in this particular case rehash is not implemented.
}

uint64_t HashTable::fill(sizetype _amount,keytype* _keys)
{
	//will contain the total number of hashfunctioncalculations for this fill process
	uint64_t maxres=0;
	//will contain the number of hashfuncitoncalculations for one insertion at a time
	uint64_t res=0;
	//insertion of the chosen elements
	for(sizetype i=0; i<_amount;i++)
	{
		res=insert(_keys[i]);
		//if rehash occured
		if(res==0)
			return res;
		else
			maxres+=res;
	}
	//calculate the average number for each insertion
	return (maxres/_amount);
}

void HashTable::print()
{
	//print the hashtables to the console formatted like:
	// cellnumber: key in table 1 | key in table 2
	for(sizetype i=0; i<p_tablesize;i++)
		std::cout<<i<<": "<<p_hashtable1[i]<<" |  "<<p_hashtable2[i]<<"\n";
	//print the stash to the console
	for(sizetype i=0; i<p_stashheight;i++)
		std::cout<<p_stash[i]<<"\n";
}

int HashTable::lookup(keytype _key)
{
	//calculate position in the first table
	sizetype cell = p_function->h1(_key);
	if(p_hashtable1[cell]==_key)
		return 1;	//if the key is present in the first table return 1
	else
	{
		//calculate position in the second table
		cell = p_function->h2(_key);
		if(p_hashtable2[cell]==_key)
			return 2;	//if the key is present in the second table return 2
		else
		{	
			//check all elements of the stash if they equal the _key
			for(sizetype i=0; i<p_stashheight;i++)
				if(p_stash[i]==_key)
					return 3; //if the key is present in the stash return 3
		}
	}
	//if the key isn't present return 0
	return 0;
}	
