#ifndef RAND_SINGLETON_H_
#define RAND_SINGLETON_H_

#include<boost/random.hpp>
#include<stdint.h>

/** Randomgenerator Singleton Class.
 * This class provides acces to a Boost Library based randomgenerator.
 * It is used to access the same Generator from various parts in the code.
 */

class RandomGenerator
{
public:
	/**Function to get a pointer to this class.
	 * If the random generator is used for the first time this function creates a new instance of the RandomGenerator and returns the pointer to this instance.
	 * Otherwise it returns the pointer to the allready existing instance of the Random Generator.
	 */
	static RandomGenerator* get(){
		if(gen==NULL)
			gen = new RandomGenerator();
		return gen;
	};

	/** Deletes the current instance of the RandomGenerator.
	 * If a instance of the Random Generator was created it deletes this instance.
	 */
	static void free()
	{
		if(gen!=NULL)
			delete gen;
		gen = NULL;
	};

public:
	/** Chooses a random number.
	 * Chooses a random unsigned 64 bit integer value from the by p_min and p_max defined range
	 */
	inline uint64_t rand(){return (*next)();};
	
	/** Chooses a random number.
	 * Chooses a random unsigned 32 bit integer value from the  by p_min and p_max defined range. Doesn't check if range exceeds uint32_t highbytes are just ignored.
	 */
	inline uint32_t rand32(){return (uint32_t)((*next)());};

	/** Changes the range the random numbers are choosen from.
	 * Changes the range the random numbers are choosen from to the range [_min .. _max]
	 * @param _min the minimal value that can be choosen by the random generator
	 * @param _max the maximal value that can be choosen by the random generator
	 */
	void range(uint64_t _min, uint64_t _max);

	/**Reseeds the randomgenerator
	 * Randomgenerator is initialized with seed time(0) by the constructor. For this reason in most cases it's not necessary to change the seed.
	 */
	void seed(uint32_t _seed);
private:
	/**
	 * The Boost Library based random generator.
	 */
	boost::variate_generator<boost::mt19937_64&, boost::uniform_int<uint64_t> >* next;
	
	/**
	 * A Merssenne Twister provided by Boost Library 
	 */
	boost::mt19937_64 p_generator;
	
	/**
	 * Min and max values for random generation
	 */
	uint64_t p_max, p_min;

private:
	static RandomGenerator* gen;
	RandomGenerator();
	~RandomGenerator();
	
};

#endif
